from __future__ import annotations

import os
from pathlib import Path
from typing import Literal, get_args, TypeAlias

import pandas as pd
from sklearn.preprocessing import OrdinalEncoder, MinMaxScaler

from protree.data.utils import BinaryScaler
from protree.meta import RANDOM_SEED
from protree.transformations import MultilabelHotEncoder

DEFAULT_DATA_DIR = os.environ.get("DEFAULT_DATA_DIR", "./data")

TStationaryDataset: TypeAlias = Literal["breast_cancer", "caltech", "compass", "diabetes", "mnist", "rhc"]
categorical_columns: dict[TStationaryDataset, list[str]] = {
    "breast_cancer": [],
    "caltech": [],
    "compass": ["age_cat", "priors_count", "c_charge_degree"],
    "diabetes": [],
    "mnist": [],
    "rhc": ["ca", "death", "sex", "dth30", "swang1", "dnr1", "ninsclas", "resp", "card", "neuro", "gastr", "renal", "meta",
            "hema", "seps", "trauma", "ortho", "race", "income"],
}


class StationaryDataset:
    def __init__(
            self,
            name: TStationaryDataset,
            directory: str = DEFAULT_DATA_DIR,
            encode_categorical_variables: bool = True,
            normalise: bool = False,
            lazy: bool = True
    ) -> None:
        """Dataset is a class designed to handle datasets for machine learning tasks. It provides functionalities for loading,
         transforming, and accessing training, validation, and test datasets.

        :param name: The name of the dataset.
        :param directory: The directory where the dataset files are located. Defaults to the DEFAULT_DATA_DIR.
        :param encode_categorical_variables: Whether to encode categorical variables. Defaults to True.
        :param lazy: Whether to lazy load the datasets. If True, the datasets are loaded only when accessed. Defaults to True.
        """
        self.name = name
        self.directory = Path(directory)
        self.encode_categorical_variables = encode_categorical_variables
        self.normalise = normalise
        self.lazy = lazy

        self._x_encoder = None
        self._x_binary_encoder = None
        self._y_encoder = None

        self._x_scaler = None

        sub_datasets = ["train", "valid", "test"]
        for ds in sub_datasets:
            setattr(self, f"_{ds}", None)

        if not lazy:
            for ds in sub_datasets:
                getattr(self, ds)

    def _x_transform(self, x: pd.DataFrame, y: pd.DataFrame) -> pd.DataFrame:
        columns = x.columns
        if self.encode_categorical_variables and categorical_columns[self.name]:
            if not self._x_encoder:
                self._x_encoder = OrdinalEncoder(
                    handle_unknown="use_encoded_value",
                    unknown_value=-1,
                    encoded_missing_value=-1,
                    dtype=int
                ).fit(x)
            x = pd.DataFrame(
                data=self._x_encoder.transform(x),
                columns=columns
            )
        if self.normalise:
            if not self._x_scaler:
                self._x_scaler = MinMaxScaler().fit(x)
            if not self._x_binary_encoder:
                self._x_binary_encoder = BinaryScaler().fit(x, y)
            x = pd.DataFrame(data=self._x_scaler.transform(x), columns=columns)
            x = pd.DataFrame(data=self._x_binary_encoder.transform(x), columns=columns)
        return x

    def _y_transform(self, y: pd.DataFrame) -> pd.DataFrame:
        if y.shape[1] > 1:
            if not self._y_encoder:
                self._y_encoder = MultilabelHotEncoder().fit(y)
            y = self._y_encoder.transform(y)
        return y

    def _read_file(self, dataset: str) -> pd.DataFrame:
        return pd.read_csv(self.directory / f"{self.name}_{dataset}.csv", index_col=[0])

    @staticmethod
    def x_y_split(df: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
        x = df[[col for col in df.columns if "target" not in col]]
        y = df[[col for col in df.columns if "target" in col]]
        if isinstance(y, pd.Series):
            y = pd.DataFrame({"target": y})
        return x, y

    def _get_x_y(self, dataset: Literal["train", "valid", "test"]) -> tuple[pd.DataFrame, pd.DataFrame]:
        x, y = StationaryDataset.x_y_split(self._read_file(dataset))
        x = self._x_transform(x, y)
        y = self._y_transform(y)
        return x, y

    @property
    def train(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Training dataset."""
        if not self._train:
            self._train = self._get_x_y("train")
        return self._train

    @property
    def valid(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Validation dataset."""
        if not self._valid:
            self._valid = self._get_x_y("valid")
        return self._valid

    @property
    def test(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Testing dataset."""
        if not self._test:
            self._test = self._get_x_y("test")
        return self._test

    @property
    def x_cols(self) -> list[str]:
        """List of feature columns"""
        return self.train[0].columns.tolist()

    @property
    def y_cols(self) -> list[str]:
        """List of label columns"""
        return self.train[1].columns.tolist()


def _save_dataframe(
        df: pd.DataFrame,
        dataset_name: str,
        directory: str = DEFAULT_DATA_DIR
) -> None:
    path = Path(directory) / f"{dataset_name}.csv"
    df.reset_index(drop=True).to_csv(path)


def _split_dataframe_and_save(
        df: pd.DataFrame,
        dataset_name: str,
        directory: str = DEFAULT_DATA_DIR,
        train_size: float = 0.6,
        valid_size: float = 0.2,
        test_size: float = 0.2
) -> None:
    if train_size + valid_size + test_size != 1.0:
        raise ValueError("Sum of train, valid, and test sizes have to sum up to 1.0")

    from sklearn.model_selection import train_test_split

    y_column = [col for col in df.columns if "target" in col][0]

    train, valid_test = train_test_split(
        df,
        stratify=df[[y_column]],
        train_size=train_size,
        random_state=RANDOM_SEED
    )
    valid, test = train_test_split(
        valid_test,
        stratify=valid_test[[y_column]],
        train_size=valid_size / (valid_size + test_size),
        random_state=RANDOM_SEED
    )

    _save_dataframe(df=train, dataset_name=f"{dataset_name}_train", directory=directory)
    _save_dataframe(df=valid, dataset_name=f"{dataset_name}_valid", directory=directory)
    _save_dataframe(df=test, dataset_name=f"{dataset_name}_test", directory=directory)


def _download_diabetes(
        directory: str = DEFAULT_DATA_DIR
) -> None:
    from io import StringIO
    from requests import get

    response = get(
        url="https://raw.githubusercontent.com/plotly/datasets/master/diabetes.csv",
        params={
            "downloadformat": "csv"
        }
    )

    df = pd.read_csv(StringIO(response.content.decode("utf-8")))
    df = df.rename(columns={
        "Outcome": "target"
    })
    _split_dataframe_and_save(df=df, dataset_name="diabetes", directory=directory)


def _download_breast_cancer(
        directory: str = DEFAULT_DATA_DIR
) -> None:
    from sklearn.datasets import load_breast_cancer

    breast_cancer = load_breast_cancer(as_frame=True)
    _split_dataframe_and_save(df=breast_cancer.frame, dataset_name="breast_cancer", directory=directory)


def _download_rhc(
        directory: str = DEFAULT_DATA_DIR
) -> None:
    from io import StringIO
    from requests import get

    response = get(
        url="https://hbiostat.org/data/repo/rhc.csv",
        params={
            "downloadformat": "csv"
        }
    )

    df = pd.read_csv(StringIO(response.content.decode("utf-8")), index_col=[0])
    df.insert(df.shape[1] - 1, "target", df.pop("cat1"))
    _split_dataframe_and_save(df=df, dataset_name="rhc", directory=directory)


def _download_compass(
        directory: str = DEFAULT_DATA_DIR
) -> None:
    from io import StringIO
    from requests import get

    response = get(
        url="https://raw.githubusercontent.com/propublica/compas-analysis/master/compas-scores-two-years.csv",
        params={
            "downloadformat": "csv"
        }
    )
    df = pd.read_csv(StringIO(response.content.decode("utf-8")))
    df = df[["age", "c_charge_degree", "race", "age_cat", "score_text", "sex", "priors_count", "days_b_screening_arrest",
             "decile_score", "is_recid", "two_year_recid", "c_jail_in", "c_jail_out"]]

    ix = (df["days_b_screening_arrest"] <= 30 & (df["days_b_screening_arrest"] >= -30) & (df["is_recid"] != -1) &
          (df["c_charge_degree"] != "O") & (df["score_text"] != "N/A"))
    df = df.loc[ix]
    df["length_of_stay"] = (pd.to_datetime(df["c_jail_out"]) - pd.to_datetime(df["c_jail_in"])).apply(lambda x: x.days)

    df_cut = df.loc[~df["race"].isin(["Native American", "Hispanic", "Asian", "Other"]), :]
    df_cut_q = df_cut[["sex", "race", "age_cat", "c_charge_degree", "score_text", "priors_count", "is_recid",
                       "two_year_recid"]].copy()
    df_cut_q["priors_count"] = df_cut_q["priors_count"].apply(
        lambda x: "0" if x <= 0 else ("1 to 3" if 1 <= x <= 3 else "More than 3"))
    df_cut_q["score_text"] = df_cut_q["score_text"].apply(lambda x: "MediumHigh" if (x == "High") | (x == "Medium") else x)
    df_cut_q["age_cat"] = df_cut_q["age_cat"].replace({"25 - 45": "25 to 45"})
    df_cut_q["sex"] = df_cut_q["sex"].replace({"Female": 1.0, "Male": 0.0})
    df_cut_q["race"] = df_cut_q["race"].apply(lambda x: 1.0 if x == "Caucasian" else 0.0)

    df = df_cut_q[["two_year_recid", "sex", "race", "age_cat", "priors_count", "c_charge_degree"]].rename(columns={
        "two_year_recid": "target"
    })
    _split_dataframe_and_save(df=df, dataset_name="compass", directory=directory)


def _download_mnist(
        directory: str = DEFAULT_DATA_DIR
) -> None:
    import gzip
    from tempfile import TemporaryDirectory

    from mnist import MNIST
    from requests import get
    from sklearn.model_selection import train_test_split

    train = ("train-images-idx3-ubyte.gz", "train-labels-idx1-ubyte.gz")
    test = ("t10k-images-idx3-ubyte.gz", "t10k-labels-idx1-ubyte.gz")

    with TemporaryDirectory() as _tmp_dir:
        tmp_dir = Path(_tmp_dir)
        for f in train + test:
            response = get(
                url=f"http://yann.lecun.com/exdb/mnist/{f}",
            )
            with (tmp_dir / f).open("wb") as fp:
                fp.write(response.content)
            with gzip.open((tmp_dir / f)) as gzip_fp:
                with (tmp_dir / f).with_suffix("").open("wb") as bin_fp:
                    bin_fp.write(gzip_fp.read())

        mnist = MNIST(_tmp_dir)

        test = mnist.load_testing()
        test_df = pd.DataFrame(
            data=[pixels + [label] for pixels, label in zip(test[0], test[1]) if label in (4, 9)],
            columns=[f"x{i}" for i in range(784)] + ["target"]
        )
        _save_dataframe(df=test_df, dataset_name="mnist_test", directory=directory)

        train_valid = mnist.load_training()
        train_valid_df = pd.DataFrame(
            data=[pixels + [label] for pixels, label in zip(train_valid[0], train_valid[1]) if label in (4, 9)],
            columns=[f"x{i}" for i in range(784)] + ["target"]
        )

        train, valid = train_test_split(
            train_valid_df,
            stratify=train_valid_df[["target"]],
            test_size=test_df.shape[0],
            random_state=RANDOM_SEED
        )
        _save_dataframe(df=train, dataset_name="mnist_train", directory=directory)
        _save_dataframe(df=valid, dataset_name="mnist_valid", directory=directory)


def _download_caltech(
        directory: str = DEFAULT_DATA_DIR
) -> None:
    from tempfile import TemporaryDirectory

    import torch
    from torchvision import transforms
    from torchvision.datasets import Caltech256
    from torchvision.models.resnet import resnet50, ResNet50_Weights

    def collate_fn(batch: list[tuple[torch.Tensor, int]]):
        images = []
        labels = []
        for idx, (image, label) in enumerate(batch):
            images.append(image.repeat(3, 1, 1) if image.shape[0] == 1 else image)
            labels.append(label)
        return torch.stack(images, dim=0), labels

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    transform = transforms.Compose([
        transforms.Resize((224, 224)),
        transforms.ToTensor()
    ])

    resnet = resnet50(ResNet50_Weights.IMAGENET1K_V2).to(device)

    with TemporaryDirectory() as _tmp_dir:
        caltech = Caltech256(root=_tmp_dir, transform=transform, download=True)
        dataloader = torch.utils.data.DataLoader(
            caltech,
            batch_size=32,
            shuffle=False,
            collate_fn=collate_fn
        )

        x = []
        y = []

        for i, (images, labels) in enumerate(dataloader):
            vector = resnet(images.to(device))
            x.extend(vector.cpu().tolist())
            y.extend(labels)

        df = pd.DataFrame(data=x, columns=[f"x{i + 1}" for i in range(len(x[0]))])
        df["target"] = y
    df = df[df["target"].isin((62, 135))]
    _split_dataframe_and_save(df=df, dataset_name="caltech", directory=directory)


def download_all(
        directory: str = DEFAULT_DATA_DIR,
        dataset_names: list[Literal["all"] | TStationaryDataset] = "all",
        verbose: bool = True
) -> None:
    path = Path(directory)
    path.mkdir(parents=True, exist_ok=True)

    allowable_datasets = get_args(TStationaryDataset)

    if "all" in dataset_names:
        dataset_names = allowable_datasets

    for dataset_name in dataset_names:
        if dataset_name not in allowable_datasets:
            raise NameError(f"Unknown dataset: {dataset_name}")
        if verbose:
            print(f"Processing {dataset_name}...")
        globals()[f"_download_{dataset_name}"](directory=directory)



from __future__ import annotations

import numpy as np
import pandas as pd


class MultilabelHotEncoder:
    def __init__(self) -> None:
        """MultilabelHotEncoder is a class designed to hot encode a DataFrame with multiple columns containing categorical
        labels. It supports fitting to the unique labels in the data and transforming the data into a binary-encoded format.

        Example:
        >>> encoder = MultilabelHotEncoder()
        >>> data = pd.DataFrame({'target1': ['foo', 'foo', 'bar'],
                                 'target2': [None, 'bar', None]})
        >>> encoded_data = encoder.fit_transform(data)
        >>> print(encoded_data)
           foo  bar
        0    1    0
        1    1    1
        2    0    1
        """
        self.columns = set()

    def fit(self, y: pd.DataFrame) -> MultilabelHotEncoder:
        """Fit the encoder to the unique labels in the input DataFrame.

        :param y: The input DataFrame with categorical labels.

        :return: The fitted encoder instance.
        """
        for col in y:
            self.columns.update(set(y[col].unique()))
        self.columns.remove(np.nan)
        self.columns = list(self.columns)
        return self

    def transform(self, y: pd.DataFrame) -> pd.DataFrame:
        """Transform the input DataFrame into a binary-encoded format.

        :param y: The input DataFrame with categorical labels.

        :return: A new DataFrame with binary-encoded columns.
        """
        y_transformed = pd.DataFrame(
            data=np.zeros((y.shape[0], len(self.columns)), dtype=int),
            columns=self.columns
        )
        for col in self.columns:
            y_transformed[col] = (y == col).any(axis=1).astype(int)
        return y_transformed

    def fit_transform(self, y: pd.DataFrame) -> pd.DataFrame:
        """
        Fit the encoder and transform the input DataFrame in a single step.

        :param y: The input DataFrame with categorical labels.

        :return: A new DataFrame with binary-encoded columns.
        """
        self.fit(y)
        return self.transform(y)



from __future__ import annotations

from typing import Literal

import numpy as np
from river.datasets import synth

from protree.data.stream_generators import IStreamGenerator


class SEA(IStreamGenerator):
    """SEA synthetic dataset, based on the River library implementation:
    https://riverml.xyz/latest/api/datasets/synth/SEA/

    The simulated drift is a transition between two classification functions available in the dataset.

    :param drift_position: The position of the drift.
    :param drift_duration: The width of the drift.
    :param seed: Random seed.
    """

    def __init__(self, drift_position: int | list[int] = 500, drift_duration: int = 0, seed: int = 42) -> None:
        super().__init__(drift_position, drift_duration, seed)
        self.drift_position = drift_position if isinstance(drift_position, (list, tuple)) else [drift_position]
        self.drift_duration = drift_duration
        self.seed = seed
        self._iter_counter = 0
        self._iter_drift_remaining = 0
        self._data_streams = (synth.SEA(variant=3, seed=seed).__iter__(), None)
        np.random.seed(seed)

    def update_drift(self) -> None:
        if self._iter_counter in self.drift_position:
            self._data_streams = (
                self._data_streams[0],
                synth.SEA(variant=np.random.randint(0, 4), seed=self.seed).__iter__()
            )
            self._iter_drift_remaining = self.drift_duration
        if self._iter_drift_remaining == 1:
            self._data_streams = self._data_streams[1], None
        self._iter_drift_remaining -= 1

    def _normalise(self, x: dict[str, float]) -> dict[str, float]:
        return {k: v / 10 for k, v in x.items()}

    def take(self, n: int) -> list[tuple[dict[str, float], int]]:
        result = []
        for i in range(n):
            self.update_drift()
            if self._iter_drift_remaining > 0 and np.random.uniform(0, 1) < self.drift_factor():
                x, y = next(self._data_streams[1])
            else:
                x, y = next(self._data_streams[0])
            x = self._normalise(x)
            result.append((x, y))
        return result


class RBF(IStreamGenerator):
    def __init__(self, drift_position: int | list[int] = 500, drift_duration: int = 1, seed: int = 42,
                 n_informative: int = 5, n_centroids: int = 11) -> None:
        super().__init__(drift_position, drift_duration, seed)
        self._data_stream = synth.RandomRBF(seed_model=seed, seed_sample=seed, n_features=n_informative,
                                            n_centroids=n_centroids).__iter__()
        self._flip = False

    def _permute_y(self, y: int) -> int:
        map_ = {0: 2, 1: 3, 2: 0, 3: 1}
        return map_[y]

    def take(self, n: int) -> list[tuple[dict[str, float], int]]:
        result = []
        for i in range(n):
            self.update_drift()
            x, y = next(self._data_stream)
            if self._flip:
                y = self._permute_y(y)
            if self._iter_drift_remaining > 0 and np.random.uniform(0, 1) < self.drift_factor():
                y = self._permute_y(y)

            result.append((x, y))
            self._iter_counter += 1
        return result

    def update_drift(self):
        if self._iter_counter in self.drift_position:
            self._iter_drift_remaining = self.drift_duration
        if self._iter_drift_remaining == 1:
            self._flip = not self._flip
        self._iter_drift_remaining -= 1


class Stagger(IStreamGenerator):
    def __init__(self, drift_position: int | list[int] = 500, drift_duration: int = 1,
                 classification_function: Literal[0, 1, 2] = 0, seed: int = 42) -> None:
        super().__init__(drift_position, drift_duration, seed)
        self._data_stream = synth.STAGGER(classification_function=classification_function, balance_classes=True,
                                          seed=seed)
        self._data_stream_iter = self._data_stream.__iter__()

    def _normalise(self, x: dict[str, float]) -> dict[str, float]:
        return {
            **{f"size_{i}": int(i == x["size"]) for i in range(3)},
            **{f"color_{i}": int(i == x["color"]) for i in range(3)},
            **{f"shape_{i}": int(i == x["shape"]) for i in range(3)},
        }

    def take(self, n: int) -> list[tuple[dict[str, float], int]]:
        result = []
        for i in range(n):
            self.update_drift()
            x, y = next(self._data_stream_iter)
            x = self._normalise(x)
            result.append((x, y))
            self._iter_counter += 1
        return result

    def update_drift(self):
        if self._iter_counter in self.drift_position:
            self._data_stream.generate_drift()
            self._iter_drift_remaining = self.drift_duration
        self._iter_drift_remaining -= 1


class Mixed(IStreamGenerator):
    def __init__(self, drift_position: int | list[int] = 500, drift_duration: int = 1,
                 classification_function: Literal[0, 1] = 1, seed: int = 42) -> None:
        super().__init__(drift_position, drift_duration, seed)
        self._data_stream = synth.Mixed(classification_function=classification_function, balance_classes=False, seed=seed)
        self._data_stream_iter = self._data_stream.__iter__()

    def _normalise(self, x: dict[int, float]) -> dict[int, float]:
        x[0] = int(x[0])
        x[1] = int(x[1])
        return x

    def take(self, n: int) -> list[tuple[dict[int, float], int]]:
        result = []
        for i in range(n):
            self.update_drift()
            x, y = next(self._data_stream_iter)
            x = self._normalise(x)
            result.append((x, y))
            self._iter_counter += 1
        return result

    def update_drift(self):
        if self._iter_counter in self.drift_position:
            self._data_stream.generate_drift()
            self._iter_drift_remaining = self.drift_duration
        self._iter_drift_remaining -= 1



from typing import Any

import numpy as np
import pandas as pd

from protree import TDataBatch, TTarget, TPrototypes


def pprint_dict(dict_: dict, indent_level: int = 0) -> None:
    indent = "\t" * indent_level
    for key, value in dict_.items():
        if isinstance(value, dict):
            print(f"{indent}{key}:")
            pprint_dict(value, indent_level + 1)
        else:
            print(f"{indent}{key}: {value}")


def parse_int_float_str(value) -> int | float | str:
    try:
        return int(value)
    except:
        pass

    try:
        return float(value)
    except:
        pass

    return value


def get_x_belonging_to_cls(x: TDataBatch, y: TTarget, cls: int | str) -> TDataBatch:
    if isinstance(x, pd.DataFrame) and isinstance(y, (np.ndarray, pd.DataFrame)):
        if isinstance(y, np.ndarray):
            return x[y == cls]
        return x[(y == cls).any(axis=1)]
    if isinstance(x, (list, tuple)) and isinstance(y, (np.ndarray, list, tuple)):
        return [x_ for x_, y_ in zip(x, y) if y_ == cls]
    raise ValueError("x and y have to both be of the same type, one of pd.DataFrame or list")


def get_x_not_belonging_to_cls(x: TDataBatch, y: TTarget, cls: int | str) -> TDataBatch:
    if isinstance(x, pd.DataFrame) and isinstance(y, (np.ndarray, pd.DataFrame)):
        if isinstance(y, np.ndarray):
            return x[(y != cls)]
        return x[(y != cls).any(axis=1)]
    if isinstance(x, (list, tuple)) and isinstance(y, (list, tuple)):
        return [x_ for x_, y_ in zip(x, y) if y_ != cls]
    raise ValueError("x and y have to both be of the same type, one of pd.DataFrame or list")


def iloc(x: pd.DataFrame | list[Any], indices: list[int]) -> pd.DataFrame | list[Any]:
    if isinstance(x, pd.DataFrame):
        return x.iloc[indices, :]
    if isinstance(x, (list, tuple)):
        return [x_ for i, x_ in enumerate(x) if i in indices]


def parse_batch(batch: TDataBatch) -> pd.DataFrame:
    if isinstance(batch, pd.DataFrame):
        return batch
    return pd.DataFrame.from_records(batch)


def parse_prototypes(prototypes: TPrototypes) -> dict[int | str, pd.DataFrame]:
    if isinstance(list(prototypes.values())[0], pd.DataFrame):
        return prototypes
    return {cls: pd.DataFrame.from_records(prototypes[cls]) for cls in prototypes}


def get_re_idx(prototypes: TPrototypes, cls: str | int, idx: int, in_class_only: bool = False) -> int:
    if isinstance(prototypes[cls], pd.DataFrame):
        if in_class_only:
            re_idx = prototypes[cls].index.to_list().index(idx)
        else:
            re_idx = flatten_prototypes(prototypes).index.to_list().index(idx)
    if isinstance(prototypes[cls], list):
        if in_class_only:
            re_idx = idx
        else:
            re_idx = 0
            for cls_ in prototypes:
                for idx_ in prototypes[cls_]:
                    if cls == cls_ and idx == idx_:
                        break
                    re_idx += 1
    return re_idx


def flatten_prototypes(prototypes: TPrototypes) -> pd.DataFrame | list[dict[str, int | float]]:
    if isinstance(prototypes[list(prototypes.keys())[0]], pd.DataFrame):
        return pd.concat([prototypes[cls] for cls in prototypes], ignore_index=False)



import pandas as pd


def remove_univalued_columns(prototypes: dict[int | str, pd.DataFrame]) -> dict[int | str, pd.DataFrame]:
    prototypes = prototypes.copy()
    prototypes_flat = pd.concat([prototypes[c] for c in prototypes])

    mask = prototypes_flat.nunique().values == 3
    columns_to_drop = prototypes_flat.columns[mask]
    for key in prototypes:
        prototypes[key] = prototypes[key].drop(columns=columns_to_drop)
    return prototypes



from __future__ import annotations

import numpy as np
import pandas as pd
from scipy.stats import entropy


class BinaryScaler:
    def __init__(
            self
    ) -> None:
        self.mapper: dict[str, float] = {}

    def fit(self, x: pd.DataFrame, y: pd.DataFrame) -> BinaryScaler:
        binary_columns = x.columns[x.nunique() == 2]
        for c in binary_columns:
            conditional_entropies = []
            for cls in y["target"].unique():
                mask = y["target"] == cls
                p_x_given_y = x[mask].loc[:, c].values.mean()
                entropy_x_given_y = entropy([p_x_given_y, 1 - p_x_given_y])
                weight = mask.sum() / len(y)
                conditional_entropies.append(entropy_x_given_y * weight)
            self.mapper[c] = (1 - sum(conditional_entropies)) / np.log(y["target"].nunique())
        return self

    def transform(self, x: pd.DataFrame) -> pd.DataFrame:
        for c in self.mapper:
            x[c] *= self.mapper[c]
        return x

    def fit_transform(self, x: pd.DataFrame, y: pd.DataFrame) -> pd.DataFrame:
        self.fit(x, y)
        return self.transform(x)



from typing import TypeAlias, Literal

from protree.data.river_generators import SEA, RBF, Mixed, Stagger
from protree.data.stream_generators import Sine, Plane, RandomTree

TNamedStream: TypeAlias = Literal[
    "sine1", "sine500", "plane100", "plane1000", "random_tree20", "random_tree500", "rbf1", "sea1", "stagger1", "mixed1"]


class INamedStream:
    pass


class NamedStreamGeneratorFactory:
    @staticmethod
    def create(name: TNamedStream) -> INamedStream:
        name = "".join([n.capitalize() for n in name.split("_")])
        return globals()[f"{name}"]()


class Sine1(Sine, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[42900, 55500, 67200, 81100], drift_duration=1, seed=42, informative_attrs=(3, 2))


class Sine500(Sine, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[41300, 60000, 73000], drift_duration=500, seed=42, informative_attrs=(3, 2))


class Plane100(Plane, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[55500, 77200, 83500, 94000], drift_duration=100, seed=42)


class Plane1000(Plane, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[48500, 72200, 88700], drift_duration=1000, seed=42)


class RandomTree20(RandomTree, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[42100, 49800, 79500], drift_duration=20, seed=42, n_informative=4, n_redundant=2)


class RandomTree500(RandomTree, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[50000, 70000, 90000], drift_duration=500, seed=42, n_informative=2, n_redundant=1)


class Rbf1(RBF, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[47100, 63000, 88800], drift_duration=1, seed=42, n_informative=4, n_centroids=9)


class Sea1(SEA, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[43900, 65000, 77500, 90000], drift_duration=1, seed=42)


class Mixed1(Mixed, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[49900], classification_function=1, drift_duration=1, seed=42)


class Stagger1(Stagger, INamedStream):
    def __init__(self):
        super().__init__(drift_position=[43500, 69500], classification_function=0, drift_duration=1, seed=42)



from __future__ import annotations

import click

from protree.data.static import download_all, DEFAULT_DATA_DIR


@click.command()
@click.option("--directory", "-d", default=DEFAULT_DATA_DIR, help="Directory to store datasets")
@click.option("--silent", "-s", is_flag=True, help="Suppress displaying progress.")
@click.option("--dataset-names", "-n", default="all", help="Comma-separated list of dataset names to download. "
                                                           "Allowable values are 'breast_cancer', 'caltech', 'compass', "
                                                           "'diabetes', 'mnist' and 'rhc'. Use 'all' to download all "
                                                           "datasets.")
def main(directory, silent, dataset_names):
    download_all(
        directory=directory,
        dataset_names=[s.strip() for s in dataset_names.split(",")],
        verbose=not silent
    )


if __name__ == "__main__":
    main()



from copy import deepcopy

import numpy as np
import pandas as pd

from protree import TPrototypes, TTarget, TDataBatch
from protree.explainers.tree_distance import IExplainer
from protree.utils import get_x_belonging_to_cls, get_re_idx, flatten_prototypes, get_x_not_belonging_to_cls


def individual_contribution(
        prototypes: TPrototypes,
        cls: str | int,
        idx: int,
        explainer: IExplainer,
        x: TDataBatch
) -> float:
    from protree.metrics.group import fidelity_with_model

    prototypes_without = deepcopy(prototypes)
    if isinstance(prototypes[cls], pd.DataFrame):
        prototypes_without[cls].drop(idx, inplace=True)
    else:
        prototypes_without[cls].pop(idx)

    result = fidelity_with_model(prototypes, explainer, x) - fidelity_with_model(prototypes_without, explainer, x)
    return result


def voting_frequency(
        prototypes: TPrototypes,
        cls: str | int,
        idx: int,
        explainer: IExplainer,
        x: TDataBatch
) -> float:
    prototypes_flat = flatten_prototypes(prototypes)
    re_idx = get_re_idx(prototypes, cls, idx)

    x_leaves = explainer.model.get_leave_indices(x)
    proto_leaves = explainer.model.get_leave_indices(prototypes_flat)
    neighbourhood = []
    for i in range(len(proto_leaves)):
        neighbourhood.append((x_leaves == proto_leaves[i]).sum(axis=1))
    return (np.vstack(neighbourhood).argmax(axis=0) == re_idx).sum().item() / len(x) if len(x) > 0 else 0.0


def consistent_votes(
        prototypes: TPrototypes,
        cls: str | int,
        idx: int,
        explainer: IExplainer,
        x: TDataBatch
) -> float:
    re_idx = get_re_idx(prototypes, cls, idx)
    prototypes_flat = flatten_prototypes(prototypes)

    classification = explainer.model.get_model_predictions(x)
    mask = classification == cls

    x_leaves = explainer.model.get_leave_indices(x)
    proto_leaves = explainer.model.get_leave_indices(prototypes_flat)
    neighbourhood = []
    for i in range(len(proto_leaves)):
        neighbourhood.append((x_leaves == proto_leaves[i]).sum(axis=1))
    votes = (np.vstack(neighbourhood).argmax(axis=0) == re_idx)
    correct = np.logical_and(votes, mask).sum()
    return (correct / votes.sum()).item() if votes.sum() > 0 else 0.0


def hubness(
        prototypes: TPrototypes,
        cls: str | int,
        idx: int,
        explainer: IExplainer,
        x: TDataBatch,
        y: TTarget
) -> float:
    sub_x = get_x_belonging_to_cls(x, y, cls)

    if not len(sub_x):
        return 0.0

    re_idx = get_re_idx(prototypes, cls, idx, in_class_only=True)

    x_cls_leaves = explainer.model.get_leave_indices(sub_x)
    proto_leaves = explainer.model.get_leave_indices(prototypes[cls])

    neighbourhood = []
    for i in range(len(proto_leaves)):
        neighbourhood.append((x_cls_leaves == proto_leaves[i]).sum(axis=1))
    return (np.vstack(neighbourhood).argmax(axis=0) == re_idx).sum().item() / len(sub_x) if len(sub_x) > 0 else 0.0


def _mean_similarity(
        sub_x: TDataBatch,
        prototypes: TPrototypes,
        cls: str | int,
        idx: int,
        explainer: IExplainer
) -> float:
    x_cls_leaves = explainer.model.get_leave_indices(sub_x)
    if isinstance(prototypes[cls], pd.DataFrame):
        proto_leaves = explainer.model.get_leave_indices(pd.DataFrame(prototypes[cls].loc[idx]).transpose())
    else:
        proto_leaves = explainer.model.get_leave_indices([prototypes[cls][idx]])
    return ((x_cls_leaves == proto_leaves).sum() / np.prod(x_cls_leaves.shape)).item()


def individual_in_distribution(
        prototypes: TPrototypes,
        cls: str | int,
        idx: int,
        explainer: IExplainer,
        x: TDataBatch,
        y: TTarget
) -> float:
    sub_x = get_x_belonging_to_cls(x, y, cls)
    return _mean_similarity(sub_x, prototypes, cls, idx, explainer)


def individual_out_distribution(
        prototypes: TPrototypes,
        cls: str | int,
        idx: int,
        explainer: IExplainer,
        x: TDataBatch,
        y: TTarget
) -> float:
    sub_x = get_x_not_belonging_to_cls(x, y, cls)
    return _mean_similarity(sub_x, prototypes, cls, idx, explainer)



from typing import TypeAlias

import numpy as np
import pandas as pd
from river.forest import ARFClassifier
from sklearn.ensemble import RandomForestClassifier as _SKLearnRandomForestClassifier

TModel: TypeAlias = ARFClassifier | _SKLearnRandomForestClassifier
TDataPoint: TypeAlias = pd.Series | np.ndarray | dict[str, int | float]
TDataBatch: TypeAlias = pd.DataFrame | np.ndarray | list[dict[str, int | float]]
TTarget: TypeAlias = pd.DataFrame | list[int | str]
TPrototypes: TypeAlias = dict[str | int, pd.DataFrame | dict[str, int | float]]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeAlias, Literal

import numpy as np
from sklearn.tree import DecisionTreeClassifier

from protree import TDataBatch, TTarget
from protree.meta import RANDOM_SEED

TStreamGenerator: TypeAlias = Literal["sine", "plane", "random_tree", "rbf"]


class StreamGeneratorFactory:
    @staticmethod
    def create(name: TStreamGenerator, drift_position: int | list[int] = 500, drift_duration: int = 5, seed: int = RANDOM_SEED
               ) -> IStreamGenerator:
        name = ''.join([n.capitalize() for n in name.split('_')])
        return globals()[f"{name}"](drift_position=drift_position, drift_duration=drift_duration, seed=seed)


class IStreamGenerator(ABC):
    def __init__(self, drift_position: int | list[int] = 500, drift_duration: int = 0, seed: int = 42) -> None:
        self.drift_position = drift_position if isinstance(drift_position, (list, tuple)) else [drift_position]
        self.drift_duration = drift_duration
        self.seed = seed
        self._iter_counter = 0
        self._iter_drift_remaining = 0
        np.random.seed(seed)

    def drift_factor(self) -> float:
        if self._iter_drift_remaining > 0:
            half_drift = self.drift_duration / 2
            return 1.0 / (1 + np.exp((2 * (-half_drift + self._iter_drift_remaining)) / half_drift))
        return 0.0

    @abstractmethod
    def take(self, n: int) -> list[tuple[TDataBatch, TTarget]]:
        pass

    def __iter__(self) -> IStreamGenerator:
        return self

    def __next__(self) -> tuple[dict[str, float], int]:
        return self.take(1)[0]


class Sine(IStreamGenerator):
    def __init__(self, drift_position: int | list[int] = 500, drift_duration: int = 1, seed: int = 42,
                 informative_attrs: tuple[int, int] = (3, 2)) -> None:
        super().__init__(drift_position, drift_duration, seed)
        self.informative_attrs_0 = informative_attrs
        self.informative_attrs_1 = (None, None)

    def take(self, n: int) -> list[tuple[dict[str, float], int]]:
        result = []
        for i in range(n):
            self.update_drift()
            x_rand = np.random.uniform(0, 1)
            x = {f"x{i}": getattr(self, f"x{i}")(x_rand) for i in range(6)}
            y = self.y(x)
            result.append((x, y))
            self._iter_counter += 1
        return result

    def _generate_new_informative_attrs(self, old_informative_attrs: tuple[int, int]) -> tuple[int, int]:
        if ((new_attrs := np.random.choice([0, 1, 2, 3], 2, replace=False)) == old_informative_attrs).all():
            return self._generate_new_informative_attrs(old_informative_attrs)
        return new_attrs

    def update_drift(self) -> None:
        if self._iter_counter in self.drift_position:
            self.informative_attrs_1 = self._generate_new_informative_attrs(self.informative_attrs_0)
            self._iter_drift_remaining = self.drift_duration
        if self._iter_drift_remaining == 1:
            self.informative_attrs_0 = self.informative_attrs_1
            self.informative_attrs_1 = (None, None)
        self._iter_drift_remaining -= 1

    def y(self, x: dict[str, float]) -> int:
        if self._iter_drift_remaining > 0 and np.random.uniform(0, 1) < self.drift_factor():
            return int(x[f"x{self.informative_attrs_1[0]}"] > x[f"x{self.informative_attrs_1[1]}"])
        return int(x[f"x{self.informative_attrs_0[0]}"] > x[f"x{self.informative_attrs_0[1]}"])

    @property
    def informative_attrs(self) -> tuple[int, int]:
        return self.informative_attrs_0

    def x0(self, x: float) -> float:
        return np.sin(np.pi * x).item()

    def x1(self, x: float) -> float:
        return (np.sin(2 * np.pi * x + np.pi / 2).item() + 1) / 2

    def x2(self, x: float) -> float:
        return np.sin(2 * np.pi / 3 * x).item()

    def x3(self, x: float) -> float:
        return (np.sin(3 * np.pi * x + np.pi / 3).item() + 1) / 2

    def x4(self, x: float) -> float:
        return np.random.uniform(0, 1)

    def x5(self, x: float) -> float:
        return np.clip(np.random.randn(1), 0, 1).item()


class Plane(IStreamGenerator):
    def __init__(self, drift_position: int | list[int] = 500, drift_duration: int = 1, seed: int = 42) -> None:
        super().__init__(drift_position, drift_duration, seed)
        self.decision_vector_0 = np.array([-0.5, 0.5])
        self.decision_vector_1 = None
        self.rotation_angle = 0
        self.direction = 1

    def take(self, n: int) -> list[tuple[dict[str, float], int]]:
        result = []
        for i in range(n):
            self.update_drift()
            x = {f"x{i}": np.random.uniform(-1, 1) for i in range(3)}
            y = self.y(x)
            result.append((x, y))
            self._iter_counter += 1
        return result

    def update_drift(self) -> None:
        if self._iter_counter in self.drift_position:
            self.direction = np.random.choice([-1, 1])
            self.decision_vector_1 = self._rotate_vector(self.decision_vector_0, self.direction * 90)
            self._iter_drift_remaining = self.drift_duration

        if self._iter_drift_remaining > 0:
            angle_per_step = self.direction * 90 / self.drift_duration
            self.rotation_angle += angle_per_step
            self.decision_vector_0 = self._rotate_vector(self.decision_vector_0, angle_per_step)

            if self._iter_drift_remaining == 1:
                self.rotation_angle = 0
                self.decision_vector_0, self.decision_vector_1 = self.decision_vector_1, None

        self._iter_drift_remaining -= 1

    def y(self, x: dict[str, float]) -> int:
        return int(self.decision_vector_0[0] * x["x0"] + self.decision_vector_0[1] * x["x1"] > 0)

    def _rotate_vector(self, vector: np.ndarray, angle_degrees: float) -> np.ndarray:
        angle_radians = np.radians(angle_degrees)
        rotation_matrix = np.array([
            [np.cos(angle_radians), -np.sin(angle_radians)],
            [np.sin(angle_radians), np.cos(angle_radians)]
        ])
        return rotation_matrix @ vector


class RandomTree(IStreamGenerator):
    def __init__(self, drift_position: int | list[int] = 500, drift_duration: int = 1, seed: int = 42,
                 n_informative: int = 4, n_redundant: int = 2) -> None:
        super().__init__(drift_position, drift_duration, seed)
        self.n_informative = n_informative
        self.n_redundant = n_redundant
        self._tree = None
        self._flip = False

    def take(self, n: int) -> list[tuple[dict[str, float], int]]:
        result = []
        for i in range(n):
            self.update_drift()
            x = {f"x{i}": np.clip(np.random.normal(loc=0.5, scale=0.2), 0, 1).item() for i in
                 range(self.n_informative + self.n_redundant)}
            y = self.y(x)
            result.append((x, y))
            self._iter_counter += 1
        return result

    def update_drift(self) -> None:
        if self._iter_counter in self.drift_position:
            self._iter_drift_remaining = self.drift_duration
        if self._iter_drift_remaining == 1:
            self._flip = not self._flip
        self._iter_drift_remaining -= 1

    def _grow_tree(self) -> None:
        self._tree = DecisionTreeClassifier(max_depth=3).fit(np.random.rand(80, self.n_informative),
                                                             np.random.randint(0, 2, 80))

    def y(self, x: dict[str, float]) -> int:
        if not self._tree:
            self._grow_tree()
        tree_prediction = self._tree.predict(np.array([list(x.values())[:self.n_informative]]))[0]
        return int(self._flip != tree_prediction)



from __future__ import annotations

from copy import deepcopy

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans as SKKMeans

from protree import TDataBatch, TTarget, TModel
from protree.explainers.tree_distance import IExplainer, ModelAdapterBuilder
from protree.explainers.utils import _type_to_np_dtype
from protree.meta import RANDOM_SEED
from protree.utils import get_x_belonging_to_cls, iloc


class KMeans(IExplainer):
    def __init__(
            self,
            model: TModel,
            n_prototypes: int = 3,
            init: str = "k-means++",
            n_init: str = "auto",
            max_iter: int = 300,
            tol: float = 0.0001,
            verbose: int = 0,
            random_state: int | None = RANDOM_SEED,
            copy_x: bool = True,
            algorithm: str = "lloyd",
            *args,
            **kwargs
    ) -> None:
        self.model = ModelAdapterBuilder(model)()
        self.prototypes = {}
        self._kmeans = {}
        self._kwargs = dict(n_clusters=n_prototypes, init=init, n_init=n_init, max_iter=max_iter, tol=tol, verbose=verbose,
                            random_state=random_state, copy_x=copy_x, algorithm=algorithm)

    def _fit_select_prototypes(self, x: TDataBatch, y: TTarget) -> None:
        classes = self.get_classes(y)
        for c in classes:
            x_partial = get_x_belonging_to_cls(x, y, c)
            x_partial_np = x_partial.values if isinstance(x_partial, pd.DataFrame) else pd.DataFrame.from_records(
                x_partial).values

            c_kwargs = self._kwargs.copy()
            c_kwargs["n_clusters"] = min(int(c_kwargs["n_clusters"]), len(x_partial_np))

            self._kmeans[c] = SKKMeans(**c_kwargs).fit(x_partial_np)
            cluster_centers = self._kmeans[c].cluster_centers_
            distances = np.linalg.norm(x_partial_np[:, np.newaxis, :] - cluster_centers[np.newaxis, :], axis=2)
            closest_indices = []
            for id_d in range(distances.shape[1]):
                closest_idx = np.argmin(distances[:, id_d])
                distances[closest_idx, :] = np.inf
                closest_indices.append(closest_idx)
            self.prototypes[c] = iloc(x_partial, closest_indices)

    def fit(self, x: pd.DataFrame) -> KMeans:
        y = pd.DataFrame({"target": self.model.get_model_predictions(x)})

        if not self.prototypes:
            self._fit_select_prototypes(x, y)
        return self

    def select_prototypes(self, x: TDataBatch, y: TTarget | None = None) -> TPrototypes:

        x = deepcopy(x)
        if not y:
            if isinstance(x, pd.DataFrame):
                y = pd.DataFrame({"target": self.model.get_model_predictions(x)})
            else:
                y = self.model.get_model_predictions(x).tolist()

        if not self.prototypes:
            self._fit_select_prototypes(x, y)
        return self.prototypes

    def predict_with_prototypes(self, x: pd.DataFrame, prototypes: dict[int | str, pd.DataFrame]) -> np.ndarray:
        predictions = (np.ones((len(x), 1)) * (-1)).astype(_type_to_np_dtype(prototypes))
        distance = np.ones((len(x))) * np.inf
        for cls in prototypes:
            for _, cluster_centers in prototypes[cls].iterrows():
                distances = np.linalg.norm(x.values[:, np.newaxis, :] - cluster_centers.values[np.newaxis, :],
                                           axis=2).squeeze()
                mask = distances < distance
                predictions[mask] = cls
                distance = np.minimum(distances, distance)
        return predictions

    def score_with_prototypes(self, x: pd.DataFrame, y: pd.DataFrame, prototypes: dict[int | str, pd.DataFrame]) -> float:
        y_hat = self.predict_with_prototypes(x, prototypes)
        return IExplainer._score(y, y_hat)



from typing import Dict, Any, Type

import numpy as np
import pandas as pd
from river.forest import ARFClassifier
from river.tree import HoeffdingTreeClassifier
from river.tree.nodes.branch import DTBranch


def _1d_np_to_2d_np(x: np.ndarray) -> np.ndarray:
    return np.expand_dims(x, axis=0)


def parse_input(x: np.ndarray | pd.Series | pd.DataFrame) -> np.ndarray:
    if isinstance(x, list) and len(x) == 1:
        return np.array(x)
    match type(x):
        case np.ndarray:
            match len(x.shape):
                case 1:
                    return _1d_np_to_2d_np(x)
                case 2:
                    return x
                case _:
                    raise ValueError("Invalid input size.")
        case pd.Series:
            return _1d_np_to_2d_np(x.to_numpy())
        case pd.DataFrame:
            return x.to_numpy()
        case _:
            raise TypeError(f"Invalid input type: {type(x)}.")


def _type_to_np_dtype(x: Dict[Any, Any]) -> Type | str:
    val = list(x.keys())[0]
    if isinstance(val, str):
        return "<U50"
    return type(val)


def predict_leaf_one(model: ARFClassifier | HoeffdingTreeClassifier, x: dict[str, float | int]) -> int:
    if isinstance(model, ARFClassifier):
        return [predict_leaf_one(model_, x) for model_ in model]

    if model._root is not None:
        if isinstance(model._root, DTBranch):
            leaf = model._root.traverse(x, until_leaf=True)
        else:
            leaf = model._root
        return id(leaf)



from typing import Literal

import numpy as np
import pandas as pd

from protree import TDataBatch, TPrototypes, TTarget
from protree.explainers.tree_distance import IExplainer
from protree.explainers.utils import _type_to_np_dtype
from protree.utils import parse_batch, parse_prototypes, flatten_prototypes


def get_euclidean_predictions(x: TDataBatch, prototypes: TPrototypes) -> np.ndarray:
    """Generate predictions based on Euclidean distance between prototypes and the input data.

    :param x: Input data.
    :type x: TDataBatch
    :param prototypes: Prototypes.
    :type prototypes: TPrototypes

    :returns: Assignment of the input data to classes based on Euclidean distance to the nearest prototype.
    :rtype: np.ndarray

    """

    prototypes = parse_prototypes(prototypes)
    x = parse_batch(x)

    predictions = (np.ones((len(x), 1)) * (-1)).astype(_type_to_np_dtype(prototypes))
    distance = np.ones((len(x))) * np.inf

    for cls in prototypes:
        if isinstance(prototypes[cls], list):
            prototypes[cls] = pd.DataFrame.from_records(prototypes[cls])
        for idx in prototypes[cls].index:
            dist = np.linalg.norm(x - prototypes[cls].loc[idx], axis=1)
            mask = dist < distance
            predictions[mask] = cls
            distance[mask] = dist[mask]
    predictions = predictions.flatten()
    return predictions


def _get_predictions(x: TDataBatch, prototypes: TPrototypes, explainer: IExplainer | None = None) -> np.ndarray:
    if explainer:
        return explainer.predict_with_prototypes(x, prototypes)
    return get_euclidean_predictions(x, prototypes)


def get_euclidean_prototype_assignment(x: TDataBatch, prototypes: TPrototypes) -> np.ndarray:
    """For each data point, finds the nearest prototype based on Euclidean distance.

    :param x: Input data.
    :type x: TDataBatch
    :param prototypes: Prototypes.
    :type prototypes: TPrototypes

    :returns: Assignment of the input data to nearest prototypes based on Euclidean distance.
    :rtype: np.ndarray

    """

    prototypes = parse_prototypes(prototypes)
    prototypes = flatten_prototypes(prototypes)
    x = parse_batch(x)

    distance_matrix = np.linalg.norm(x.values[:, None, :] - prototypes.values[None, ...], axis=-1)
    return distance_matrix.argmin(axis=1)


def _get_nearest_prototypes(x: TDataBatch, prototypes: TPrototypes, explainer: IExplainer) -> np.ndarray:
    if explainer:
        return explainer.get_prototype_assignment(x, prototypes)
    return get_euclidean_prototype_assignment(x, prototypes)


def _get_assignment(a: TPrototypes, b: TPrototypes, x: TDataBatch, explainer_a: IExplainer | None = None,
                    explainer_b: IExplainer | None = None, assign_to: Literal["prototype", "class"] = "class"
                    ) -> tuple[np.ndarray, np.ndarray]:
    if explainer_a and not explainer_b:
        explainer_b = explainer_a

    match assign_to:
        case "prototype":
            assignment_a = _get_nearest_prototypes(x, a, explainer_a)
            assignment_b = _get_nearest_prototypes(x, b, explainer_b)
        case "class":
            assignment_a = _get_predictions(x, a, explainer_a)
            assignment_b = _get_predictions(x, b, explainer_b)
        case _:
            raise ValueError(f"Invalid value for assign_to: {assign_to}")

    return assignment_a.reshape(len(assignment_a)), assignment_b.reshape(len(assignment_b))


def mutual_information(a: TPrototypes, b: TPrototypes, x: TDataBatch, explainer_a: IExplainer | None = None,
                       explainer_b: IExplainer | None = None, assign_to: Literal["prototype", "class"] = "class") -> float:
    """Calculate the similarity metric based on mutual information between two sets of prototypes, with respect to the input
    data. If an explainer is provided, the predictions are based on the explainer, otherwise, the predictions are based on
    Euclidean distance.

    :param a: First set of prototypes.
    :type a: TPrototypes
    :param b: Second set of prototypes.
    :type b: TPrototypes
    :param x: Input data.
    :type x: TDataBatch
    :param explainer_a: Explainer to generate predictions for the set of prototypes a (Optional).
    :type explainer_a: IExplainer | None
    :param explainer_b: Explainer to generate predictions for the set of predictions b. If explainer_a is provided and
    explainer_b is not provided, explainer_a is used for both prototypes a and b (Optional).
    :type explainer_b: IExplainer | None
    :param assign_to: Whether to compute mutual information based on prototype or class assignment (default = "class").
    :type assign_to: Literal["prototype", "class"]

    :returns: The similarity metric between the two sets of prototypes based on mutual information.
    :rtype: float

    """

    from sklearn.metrics import adjusted_mutual_info_score

    assignment_a, assignment_b = _get_assignment(a, b, x, explainer_a, explainer_b, assign_to)

    return adjusted_mutual_info_score(assignment_a, assignment_b, average_method="arithmetic")


def rand_index(a: TPrototypes, b: TPrototypes, x: TDataBatch, explainer_a: IExplainer | None = None,
               explainer_b: IExplainer | None = None, assign_to: Literal["prototype", "class"] = "class") -> float:
    """Calculate the similarity metric based on the Rand index between two sets of prototypes, with respect to the input
    data. If an explainer is provided, the predictions are based on the explainer, otherwise, the predictions are based on
    Euclidean distance.

    :param a: First set of prototypes.
    :type a: TPrototypes
    :param b: Second set of prototypes.
    :type b: TPrototypes
    :param x: Input data.
    :type x: TDataBatch
    :param explainer_a: Explainer to generate predictions for the set of prototypes a (Optional).
    :type explainer_a: IExplainer | None
    :param explainer_b: Explainer to generate predictions for the set of predictions b. If explainer_a is provided and
    explainer_b is not provided, explainer_a is used for both prototypes a and b (Optional).
    :type explainer_b: IExplainer | None
    :param assign_to: Whether to compute mutual information based on prototype or class assignment (default = "class").
    :type assign_to: Literal["prototype", "class"]

    :returns: The similarity metric between the two sets of prototypes based on the Rand index.
    :rtype: float

    """

    from sklearn.metrics import adjusted_rand_score

    assignment_a, assignment_b = _get_assignment(a, b, x, explainer_a, explainer_b, assign_to)

    return adjusted_rand_score(assignment_a, assignment_b)


def fowlkes_mallows(a: TPrototypes, b: TPrototypes, x: TDataBatch, explainer_a: IExplainer | None = None,
                    explainer_b: IExplainer | None = None, assign_to: Literal["prototype", "class"] = "class") -> float:
    """Calculate the similarity metric based on the Fowlkes-Mallows index between two sets of prototypes, with respect to the
    input data. If an explainer is provided, the predictions are based on the explainer, otherwise, the predictions are based
    on Euclidean distance.

    :param a: First set of prototypes.
    :type a: TPrototypes
    :param b: Second set of prototypes.
    :type b: TPrototypes
    :param x: Input data.
    :type x: TDataBatch
    :param explainer_a: Explainer to generate predictions for the set of prototypes a (Optional).
    :type explainer_a: IExplainer | None
    :param explainer_b: Explainer to generate predictions for the set of predictions b. If explainer_a is provided and
    explainer_b is not provided, explainer_a is used for both prototypes a and b (Optional).
    :type explainer_b: IExplainer | None
    :param assign_to: Whether to compute mutual information based on prototype or class assignment (default = "class").
    :type assign_to: Literal["prototype", "class"]

    :returns: The similarity metric between the two sets of prototypes based on the Fowlkes-Mallows index.
    :rtype: float

    """

    from sklearn.metrics import fowlkes_mallows_score

    if explainer_a and not explainer_b:
        explainer_b = explainer_a

    assignment_a, assignment_b = _get_assignment(a, b, x, explainer_a, explainer_b, assign_to)

    return fowlkes_mallows_score(assignment_a, assignment_b)


def completeness(a: TPrototypes, b: TPrototypes, x: TDataBatch, explainer_a: IExplainer | None = None,
                 explainer_b: IExplainer | None = None, assign_to: Literal["prototype", "class"] = "class") -> float:
    """Calculate the similarity metric based on the completeness score between two sets of prototypes, with respect to the
    input data. If an explainer is provided, the predictions are based on the explainer, otherwise, the predictions are based
    on Euclidean distance.

    :param a: First set of prototypes.
    :type a: TPrototypes
    :param b: Second set of prototypes.
    :type b: TPrototypes
    :param x: Input data.
    :type x: TDataBatch
    :param explainer_a: Explainer to generate predictions for the set of prototypes a (Optional).
    :type explainer_a: IExplainer | None
    :param explainer_b: Explainer to generate predictions for the set of predictions b. If explainer_a is provided and
    explainer_b is not provided, explainer_a is used for both prototypes a and b (Optional).
    :type explainer_b: IExplainer | None
    :param assign_to: Whether to compute mutual information based on prototype or class assignment (default = "class").
    :type assign_to: Literal["prototype", "class"]

    :returns: The similarity metric between the two sets of prototypes based on the completeness score.
    :rtype: float

    """

    from sklearn.metrics import completeness_score

    if explainer_a and not explainer_b:
        explainer_b = explainer_a

    assignment_a, assignment_b = _get_assignment(a, b, x, explainer_a, explainer_b, assign_to)

    return completeness_score(assignment_a, assignment_b)


def centroids_displacements(a: TPrototypes, b: TPrototypes, penalty: float = 10.0) -> dict[int | str, float]:
    """Calculate the displacement of the centroids between two sets of prototypes.

    :param a: First set of prototypes.
    :type a: TPrototypes
    :param b: Second set of prototypes.
    :type b: TPrototypes
    :param penalty: Penalty for empty classes (default = 10.0).
    :type penalty: float
    """

    classes = set(a.keys()).union(b.keys())

    a = parse_prototypes(a)
    b = parse_prototypes(b)

    distances = {}

    for cls in classes:
        # both empty
        if (cls not in a or len(a[cls]) == 0) and (cls not in b or len(b[cls]) == 0):
            distances[cls] = 0.0

        # one empty
        elif len(a[cls]) == 0 or len(b[cls]) == 0:
            distances[cls] = penalty

        # both not empty
        else:
            distances[cls] = np.linalg.norm((a[cls].mean() - b[cls].mean()).values)
    return distances


def mean_centroid_displacement(a: TPrototypes, b: TPrototypes, penalty: float = 2.0) -> float:
    """Calculate the mean displacement of the centroids between two sets of prototypes.

    :param a: First set of prototypes.
    :type a: TPrototypes
    :param b: Second set of prototypes.
    :type b: TPrototypes
    :param penalty: Penalty for empty classes (default = 2.0).
    :type penalty: float

    :returns: The mean displacement of the centroids between the two sets of prototypes.
    :rtype: float

    """

    displacements = centroids_displacements(a, b, penalty)
    return np.mean(list(displacements.values()))


def _l2_distance_matrix(a: TPrototypes, b: TPrototypes, cls: int | float) -> np.ndarray:
    axis = int(len(a[cls]) > len(b[cls]))
    a[cls] = parse_batch(a[cls])
    b[cls] = parse_batch(b[cls])
    distance_matrix = np.linalg.norm(a[cls].values[:, None, ...] - b[cls].values[None, ...], axis=-1)
    return distance_matrix.min(axis=axis).tolist()


def mean_minimal_distance(a: TPrototypes, b: TPrototypes, penalty: float = 2.0) -> float:
    """Calculate the mean minimal distance between two sets of prototypes.

    :param a: First set of prototypes.
    :type a: TPrototypes
    :param b: Second set of prototypes.
    :type b: TPrototypes
    :param penalty: Penalty for empty classes (default = 2.0).
    :type penalty: float

    :returns: The mean minimal distance between the two sets of prototypes.
    :rtype: float

    """

    classes = set(a.keys()).union(b.keys())
    distances = []
    for cls in classes:
        # both empty
        if (cls not in a or len(a[cls]) == 0) and (cls not in b or len(b[cls]) == 0):
            continue

        # one empty
        elif cls not in a or len(a[cls]) == 0:
            distances.extend([penalty] * len(b[cls]))
        elif cls not in b or len(b[cls]) == 0:
            distances.extend([penalty] * len(a[cls]))

        # both not empty
        else:
            distances.extend(_l2_distance_matrix(a, b, cls))

    return np.mean(distances)


def classwise_mean_minimal_distance(a: TPrototypes, b: TPrototypes, penalty: float = 2.0) -> dict[int | str, float]:
    """Calculate the class-wise mean minimal distance between two sets of prototypes.

    :param a: First set of prototypes.
    :type a: TPrototypes
    :param b: Second set of prototypes.
    :type b: TPrototypes
    :param penalty: Penalty for empty classes (default = 2.0).
    :type penalty: float

    :returns: The class-wise mean minimal distance between the two sets of prototypes.
    :rtype: dict[int | str, float]

    """

    classes = set(a.keys()).union(b.keys())
    distances = {}
    for cls in classes:
        # both empty
        if (cls not in a or len(a[cls]) == 0) and (cls not in b or len(b[cls]) == 0):
            distances[cls] = 0.0

        # one empty
        elif (cls not in b or len(a[cls]) == 0) or (cls not in b or len(b[cls]) == 0):
            distances[cls] = penalty * len(b[cls] if cls in b and b[cls] else a[cls])

        # both not empty
        else:
            distances[cls] = np.mean(_l2_distance_matrix(a, b, cls))
    return distances


def _get_accuracy(prototypes: TPrototypes, x: TDataBatch, y: TTarget, explainer: IExplainer | None = None) -> float:
    """Get accuracy of predictions based on the given prototypes using the Euclidean distance or a custom explainer.

    :param prototypes: Prototypes.
    :type prototypes: TPrototypes
    :param x: Input data.
    :type x: TDataBatch
    :param y: True class labels for the input data.
    :type y: TTarget
    :param explainer: Explainer to generate predictions (Optional).
    :type explainer: IExplainer

    :returns: The accuracy of the predictions based on the given prototypes.
    :rtype: float

    """

    from sklearn.metrics import accuracy_score

    predictions = _get_predictions(x, prototypes, explainer)
    return accuracy_score(y, predictions)


def _one_way_swap_delta(prototypes_a: TPrototypes, prototypes_b: TPrototypes, x: TDataBatch, y: TTarget,
                        explainer: IExplainer | None = None) -> float:
    baseline_accuracy = _get_accuracy(prototypes_b, x, y, explainer)

    prototypes_a = parse_prototypes(prototypes_a)
    prototypes_b = parse_prototypes(prototypes_b)

    accuracy_changes = []

    for cls, prototypes in prototypes_a.items():
        if len(prototypes) == 0:
            continue

        for idx, prototype in prototypes.iterrows():
            temp_prototypes = prototypes_b.copy()
            if cls not in temp_prototypes:
                temp_prototypes[cls] = pd.DataFrame()
            temp_prototypes[cls] = pd.concat([temp_prototypes[cls], prototype.to_frame().T]).reset_index(drop=True)

            new_accuracy = _get_accuracy(temp_prototypes, x, y, explainer)
            accuracy_changes.append(np.abs(baseline_accuracy - new_accuracy))

    if accuracy_changes:
        return np.mean(accuracy_changes)
    else:
        return 0.0


def swap_delta(prototypes_a: TPrototypes, prototypes_b: TPrototypes, x: TDataBatch, y: TTarget,
               explainer: IExplainer | None = None) -> float:
    """
    Calculate a distance metric between two sets of prototypes by measuring the change in accuracy when prototypes from
    one set are temporarily added to the other set. This metrics is symmetric as it calculates changes in accuracy in both
    directions (from prototypes a to b and from prototypes b to a).

    :param prototypes_a: First set of prototypes.
    :type prototypes_a: TPrototypes
    :param prototypes_b: Second set of prototypes.
    :type prototypes_b: TPrototypes
    :param x: Input data.
    :type x: TDataBatch
    :param y: True class labels for the input data.
    :type y: TTarget
    :param explainer: Explainer to generate predictions (Optional).
    :type explainer: IExplainer

    :returns: A distance metric score, with higher values indicating greater divergence between the two sets of prototypes.
    :rtype: float
    """
    one_way_deterioration_a = _one_way_swap_delta(prototypes_a, prototypes_b, x, y, explainer)
    one_way_deterioration_b = _one_way_swap_delta(prototypes_b, prototypes_a, x, y, explainer)
    return (one_way_deterioration_a + one_way_deterioration_b) / 2



import numpy as np


def balanced_accuracy(y: np.ndarray, y_hat: np.ndarray) -> float:
    acc = 0.0
    for label in np.unique(y):
        mask = (y == label)
        acc_partial = (y[mask] == y_hat[mask]).mean()
        acc += (acc_partial * mask.sum()) / len(y)
    return acc.item()



from typing import Callable

import numpy as np
import pandas as pd

from protree import TPrototypes, TDataBatch, TTarget
from protree.explainers.tree_distance import IExplainer
from protree.metrics.classification import balanced_accuracy


def fidelity_with_model(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch) -> float:
    y_model = explainer.model.get_model_predictions(x)
    y_prototypes = explainer.predict_with_prototypes(x, prototypes)
    return balanced_accuracy(y_model, y_prototypes)


def contribution(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch) -> float:
    from protree.metrics.individual import individual_contribution

    diversity = 0
    n_prototypes = 0
    for cls in prototypes:
        for idx in (prototypes[cls].index if isinstance(prototypes[cls], pd.DataFrame) else range(len(prototypes[cls]))):
            try:
                diversity += individual_contribution(prototypes, cls, idx, explainer, x)
                n_prototypes += 1
            except ValueError:
                continue
    return diversity / n_prototypes


def _dist(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch, y: TTarget, func: Callable) -> float:
    dist = 0
    n_prototypes = 0
    for cls in prototypes:
        for idx in (prototypes[cls].index if isinstance(prototypes[cls], pd.DataFrame) else range(len(prototypes[cls]))):
            dist += func(prototypes, cls, idx, explainer, x, y)
            n_prototypes += 1
    return dist / n_prototypes


def mean_in_distribution(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch, y: TTarget) -> float:
    from protree.metrics.individual import individual_in_distribution
    return _dist(prototypes, explainer, x, y, individual_in_distribution)


def mean_out_distribution(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch, y: TTarget) -> float:
    from protree.metrics.individual import individual_out_distribution
    return _dist(prototypes, explainer, x, y, individual_out_distribution)


def mean_entropy_hubness(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch) -> float:
    from numpy import mean

    entropies = vector_entropy_hubness(prototypes, explainer, x)
    return mean(list(entropies.values())).item()


def vector_entropy_hubness(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch) -> dict[str | bool | int, float]:
    from numpy import log2
    from scipy.stats import entropy

    from protree.metrics.individual import hubness

    y = explainer.model.get_model_predictions(x)

    hub_score = {cls: [] for cls in prototypes}
    for cls in prototypes:
        for idx in (prototypes[cls].index if isinstance(prototypes[cls], pd.DataFrame) else range(len(prototypes[cls]))):
            hub_score[cls].append(hubness(prototypes, cls, idx, explainer, x, y))
    entropies = {
        cls: (entropy(hub_score[cls], base=2) / max(log2(len(hub_score[cls]) + 1e-6), 1 + 1e-6)).item()
        if sum(hub_score[cls]) != 0 else 0.0
        for cls in hub_score}
    return entropies


def vector_consistent_votes(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch) -> dict[str | bool | int, float]:
    from protree.metrics.individual import consistent_votes

    consistency = {cls: [] for cls in prototypes}
    for cls in prototypes:
        for idx in (prototypes[cls].index if isinstance(prototypes[cls], pd.DataFrame) else range(len(prototypes[cls]))):
            consistency[cls].append(consistent_votes(prototypes, cls, idx, explainer, x))
    return consistency


def vector_voting_frequency(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch) -> dict[str | bool | int, float]:
    from protree.metrics.individual import voting_frequency

    votes = {cls: [] for cls in prototypes}
    for cls in prototypes:
        for idx in (prototypes[cls].index if isinstance(prototypes[cls], pd.DataFrame) else range(len(prototypes[cls]))):
            votes[cls].append(voting_frequency(prototypes, cls, idx, explainer, x))
    return votes


def vector_in_distribution(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch, y: TTarget) -> dict[
    str | bool | int, float]:
    from protree.metrics.individual import individual_in_distribution

    in_distribution = {cls: [] for cls in prototypes}
    for cls in prototypes:
        for idx in (prototypes[cls].index if isinstance(prototypes[cls], pd.DataFrame) else range(len(prototypes[cls]))):
            in_distribution[cls].append(individual_in_distribution(prototypes, cls, idx, explainer, x, y))
    return in_distribution


def vector_out_distribution(prototypes: TPrototypes, explainer: IExplainer, x: TDataBatch, y: TTarget) -> dict[
    str | bool | int, float]:
    from protree.metrics.individual import individual_out_distribution

    out_distribution = {cls.item() if isinstance(cls, np.ndarray) else cls: [] for cls in prototypes}
    for cls in prototypes:
        for idx in (prototypes[cls].index if isinstance(prototypes[cls], pd.DataFrame) else range(len(prototypes[cls]))):
            out_distribution[cls].append(individual_out_distribution(prototypes, cls, idx, explainer, x, y))
    return out_distribution



from __future__ import annotations

from copy import deepcopy
from typing import Type, Literal, Callable

from river.base import DriftDetector
from river.forest import ARFClassifier
from sklearn.ensemble import RandomForestClassifier

from protree import TPrototypes, TDataBatch, TTarget
from protree.explainers import TExplainer, APete


class ProtoBDrift(DriftDetector):
    """Prototypes-based drift detector for block data streams.

    """

    def __init__(self, model: RandomForestClassifier | ARFClassifier, prototype_selector: Type[TExplainer] = APete,
                 prototype_selector_kwargs: dict | None = None,
                 measure: Literal[
                     "mutual_information", "rand_index", "completeness", "fowlkes_mallows", "centroid_displacement",
                     "minimal_distance"] = "centroid_displacement", strategy: Literal["class", "total"] = "total",
                 distance: Literal["l2", "tree"] = "l2", assign_to: Literal["class", "prototype"] = "prototype",
                 cold_start: int = 20, const: float = 0.33) -> None:
        super().__init__()
        self.assign_to = assign_to
        self.cold_start = cold_start
        self.const = const
        self.distance = distance
        self.measure = measure
        self.model = model
        self.prototype_selector = prototype_selector
        self.prototype_selector_kwargs = prototype_selector_kwargs or {}
        self.strategy = strategy

        self._metric_significance_max: int | dict[int | str, float] | None = None
        self._metric_significance_min: int | dict[int | str, float] | None = None

        self._direction: str | None = None
        self._iter_counter: int = 0
        self._drift_detected: bool = False

        self._init_state()

    def _init_state(self) -> None:
        self.x_blocks: tuple[TDataBatch, TDataBatch] = (None, None)
        self.y_blocks: tuple[TTarget, TTarget] = (None, None)

        self.explainers: tuple[TExplainer | None, TExplainer | None] = (None, None)

        self.prototypes: tuple[TPrototypes, TPrototypes] = ({}, {})

    def update(self, x: TDataBatch, y: TTarget) -> None:
        self._update_block(x, y)
        self._drift_detected = False
        self._update_explainers()
        self._find_prototypes()
        if self._iter_counter >= 1:
            metric = self._compute_metric()

            if (self.cold_start / 2) <= self._iter_counter < self.cold_start:
                self._update_metric_stats(metric)
            elif self.cold_start <= self._iter_counter:
                self._test(metric)
        self._iter_counter += 1

    def _update_block(self, x: TDataBatch, y: TTarget) -> None:
        self.x_blocks = (self.x_blocks[1], x)
        self.y_blocks = (self.y_blocks[1], y)

    def _update_explainers(self) -> None:
        self.explainers = (
            deepcopy(self.explainers[1]),
            self.prototype_selector(deepcopy(self.model), **self.prototype_selector_kwargs)
        )

    def _find_prototypes(self) -> None:
        self.prototypes = (
            deepcopy(self.prototypes[1]),
            self.explainers[1].select_prototypes(self.x_blocks[1], self.y_blocks[1])
        )

    def _update_metric_stats(self, metric: float | dict[str | int, float]) -> None:
        if self._metric_significance_min is None:
            if isinstance(metric, float):
                self._metric_significance_min = metric
                self._metric_significance_max = metric
            elif isinstance(metric, dict):
                self._metric_significance_min = {}
                self._metric_significance_max = {}
                for key in metric:
                    self._metric_significance_min[key] = metric
                    self._metric_significance_max[key] = metric

        elif isinstance(self._metric_significance_min, float):
            self._metric_significance_min = min(self._metric_significance_min, metric)
            self._metric_significance_max = max(self._metric_significance_max, metric)

        elif isinstance(self._metric_significance_min, dict):
            for key in metric:
                self._metric_significance_min[key] = min(self._metric_significance_min[key], metric[key])
                self._metric_significance_max[key] = max(self._metric_significance_max[key], metric[key])

    def _reset(self):
        super()._reset()
        self._init_state()

    def _test_value(self, value: float, key: int | str | None = None) -> bool:
        if key is not None:
            interval = self._metric_significance_max[key] - self._metric_significance_min[key]
            if self._direction == "increase":
                return value > self._metric_significance_max[key] + self.const * interval
            else:
                return value < self._metric_significance_min[key] - self.const * interval

        interval = self._metric_significance_max - self._metric_significance_min

        if self._direction == "increase":
            return value > self._metric_significance_max + self.const * interval
        else:
            return value < self._metric_significance_min - self.const * interval

    def _test(self, metric: float | dict[int | str, float]) -> None:
        # strategy: class
        if isinstance(metric, dict):
            for key in metric:
                if self._test_value(metric[key], key=key):
                    self._drift_detected = True
                    break

        # strategy: total
        elif isinstance(metric, float):
            if self._test_value(metric, key=None):
                self._drift_detected = True

    def _compute_metric(self) -> float | dict[int | str, float]:
        if self.measure in ["mutual_information", "rand_index", "completeness", "fowlkes_mallows"]:
            import protree.metrics.compare

            self._direction = "decrease"

            metric = getattr(protree.metrics.compare, self.measure)
            return self._compute_cluster_metric(metric)

        elif self.measure == "swap_delta":
            from protree.metrics.compare import swap_delta

            self._direction = "increase"
            if self.distance == "tree":
                return swap_delta(*self.prototypes, self.x_blocks[1], self.y_blocks[1], explainer=self.explainers[1])
            return swap_delta(*self.prototypes, self.x_blocks[1], self.y_blocks[1])

        elif self.measure == "minimal_distance":
            self._direction = "increase"

            if self.strategy == "class":
                from protree.metrics.compare import classwise_mean_minimal_distance

                return self._compute_spatial_metric(classwise_mean_minimal_distance)

            elif self.strategy == "total":
                from protree.metrics.compare import mean_minimal_distance

                return self._compute_spatial_metric(mean_minimal_distance)

        elif self.measure == "centroid_displacement":
            self._direction = "increase"

            if self.strategy == "class":
                from protree.metrics.compare import centroids_displacements

                return self._compute_spatial_metric(centroids_displacements)

            elif self.strategy == "total":
                from protree.metrics.compare import mean_centroid_displacement

                return self._compute_spatial_metric(mean_centroid_displacement)

    def _build_cluster_kwargs(self) -> dict:
        kwargs = {
            "a": self.prototypes[0],
            "b": self.prototypes[1],
            "x": self.x_blocks[1],
            "assign_to": self.assign_to
        }

        if self.distance == "tree":
            kwargs["explainer_a"] = self.explainers[0]
            kwargs["explainer_b"] = self.explainers[1]

        return kwargs

    def _compute_cluster_metric(self, metric: Callable[..., float]) -> float | dict[int | str, float]:
        return metric(**self._build_cluster_kwargs())

    def _compute_spatial_metric(self, metric: Callable[..., float | dict[int | str, float]]) -> float | dict[int | str, float]:
        return metric(*self.prototypes)



from setuptools import setup, find_packages

SHORT_DESCRIPTION = ("Protree is  a set of utilities for prototype selection in tree-based models and usage of prototypes in "
                     "drift detection.")

with open("README.md", "r", encoding="utf-8") as fp:
    LONG_DESCRIPTION = fp.read()

with open("requirements.txt", "r", encoding="utf-8") as fp:
    REQUIREMENTS = fp.read().split("\n")

with open("requirements-optional.txt", "r", encoding="utf-8") as fp:
    REQUIREMENTS_OPTIONAL = fp.read().split("\n")

setup(
    name="protree",
    version='v0.1.1',
    author="jkarolczak",
    author_email="jacek.karolczak@outlook.com",
    description=SHORT_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/jkarolczak/protree",
    packages=find_packages(where="."),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=REQUIREMENTS,
    extras_require={
        "all": REQUIREMENTS_OPTIONAL
    }
)



from __future__ import annotations

from enum import Enum
from typing import TypeAlias, Union

from protree.explainers.naive import KMeans
from protree.explainers.tree_distance import SM_A, SM_WA, SG, G_KM, APete

TExplainer: TypeAlias = Union[KMeans, G_KM, SM_A, SM_WA, SG, APete]


class Explainer(Enum):
    KMeans = KMeans
    G_KM = G_KM
    SM_A = SM_A
    SM_WA = SM_WA
    SG = SG
    APete = APete



import warnings
from abc import ABC
from copy import deepcopy
from typing import Iterable

import numpy as np
import pandas as pd
from river.forest import ARFClassifier
from sklearn.ensemble import RandomForestClassifier as _SKLearnRandomForestClassifier

from protree import TModel, TDataPoint, TDataBatch, TTarget, TPrototypes
from protree.explainers.utils import parse_input, _type_to_np_dtype, predict_leaf_one
from protree.metrics.classification import balanced_accuracy
from protree.utils import iloc, get_x_belonging_to_cls, flatten_prototypes, parse_prototypes


class IModelAdapter(ABC):
    @property
    def n_trees(self) -> int:
        pass

    def get_leave_indices(self, x: TDataBatch) -> np.ndarray:
        pass

    def get_model_predictions(self, x: TDataBatch) -> np.ndarray:
        pass


class SKLearnAdapter(IModelAdapter):
    def __init__(self, model: TModel) -> None:
        self.model = model

    @property
    def n_trees(self) -> int:
        return self.model.n_estimators

    def get_leave_indices(self, x: pd.DataFrame | np.ndarray) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return self.model.apply(parse_input(x))

    def get_model_predictions(self, x: pd.DataFrame | np.ndarray) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return self.model.predict(parse_input(x))


class RiverAdapter(IModelAdapter):
    def __init__(self, model: ARFClassifier) -> None:
        self.model = model

    @property
    def n_trees(self) -> int:
        return self.model.n_models

    def get_leave_indices(self, x: list[dict[str, int | float]]) -> np.ndarray:
        return np.array([predict_leaf_one(self.model, x_) for x_ in x])

    def get_model_predictions(self, x: list[dict[str, int | float]]) -> np.ndarray:
        return np.array([self.model.predict_one(x_) for x_ in x])


class ModelAdapterBuilder:
    def __init__(self, model: TModel) -> None:
        self.model = model

    def __call__(self, *args, **kwargs) -> IModelAdapter:
        if isinstance(self.model, _SKLearnRandomForestClassifier):
            return SKLearnAdapter(model=self.model)
        elif isinstance(self.model, ARFClassifier):
            return RiverAdapter(model=self.model)
        raise ValueError("Unsupported model class.")


class IExplainer(ABC):
    def __init__(self, model: TModel, *args, **kwargs) -> None:
        self.model = ModelAdapterBuilder(model)()

    def pair_similarity(self, x1: TDataPoint, x2: TDataPoint) -> float:
        nodes1 = self.model.get_leave_indices(parse_input(x1))
        nodes2 = self.model.get_leave_indices(parse_input(x2))

        return sum((nodes1 == nodes2)[0]) / self.model.n_trees

    def similarity_matrix(self, batch: TDataBatch) -> np.ndarray:
        nodes = self.model.get_leave_indices(batch)
        return np.dot(nodes, nodes.T) / self.model.n_trees

    def distance_matrix(self, batch: TDataBatch) -> np.ndarray:
        return 1 - self.similarity_matrix(batch)

    def _create_distance_matrices(self, x: TDataBatch, y: TTarget, classes: Iterable[int | str]
                                  ) -> dict[int | str, np.ndarray]:
        distances = {cls: None for cls in classes}
        for cls in classes:
            class_x = get_x_belonging_to_cls(x, y, cls)
            distances[cls] = self.distance_matrix(class_x)
        return distances

    def predict_with_prototypes(self, x: TDataBatch, prototypes: TPrototypes) -> np.ndarray:
        predictions = (np.ones((len(x), 1)) * (-1)).astype(_type_to_np_dtype(prototypes))
        similarity = np.zeros((len(x)))
        x_nodes = self.model.get_leave_indices(x)

        for cls in prototypes:
            for _, prototype in (
                    prototypes[cls].iterrows() if hasattr(prototypes[cls], "iterrows") else enumerate(prototypes[cls])):
                prototype_leaves = self.model.get_leave_indices([prototype])
                prototype_similarity = (prototype_leaves == x_nodes).sum(axis=1) / self.model.n_trees
                mask = prototype_similarity > similarity
                predictions[mask] = cls
                similarity[mask] = prototype_similarity[mask]
        return predictions

    def get_prototype_assignment(self, x: TDataBatch, prototypes: TPrototypes) -> np.ndarray:
        prototypes = deepcopy(prototypes)
        prototypes = parse_prototypes(prototypes)
        prototypes = flatten_prototypes(prototypes)
        x_nodes = self.model.get_leave_indices(x)

        predictions = np.ones((len(x), 1)) * (-1)
        similarity = np.zeros((len(x)))

        for idx, prototype in prototypes.iterrows():
            prototype_leaves = self.model.get_leave_indices([prototype])
            prototype_similarity = (prototype_leaves == x_nodes).sum(axis=1) / self.model.n_trees
            mask = prototype_similarity > similarity
            predictions[mask] = idx
            similarity[mask] = prototype_similarity[mask]
        return predictions

    @staticmethod
    def _score(y: pd.DataFrame, y_hat: np.ndarray) -> float:
        if isinstance(y, pd.DataFrame):
            y = y.values.squeeze(1)
        elif isinstance(y, list):
            y = np.array(y)
        return balanced_accuracy(y, y_hat)

    def score_with_prototypes(self, x: TDataBatch, y: TTarget, prototypes: TPrototypes) -> float:
        y_hat = self.predict_with_prototypes(x, prototypes)
        return IExplainer._score(y, y_hat)

    def score(self, x: pd.DataFrame, y: pd.DataFrame) -> float:
        y_hat = self.model.get_model_predictions(x)
        return IExplainer._score(y, y_hat)

    def get_classes(self, y: TTarget) -> set[int | str]:
        if isinstance(y, (list, tuple)):
            return set(y)
        if isinstance(y, pd.DataFrame):
            classes = set()
            for col in y:
                classes.update(set([y_.item() if hasattr(y_, "item") else y_ for y_ in y[col].unique()]))
            return classes
        if isinstance(y, np.ndarray):
            return set(y.tolist())


class G_KM(IExplainer):
    def __init__(self, model: TModel, n_prototypes: int = 3, *args, **kwargs) -> None:
        super().__init__(model, *args, **kwargs)
        self.n_prototypes = int(n_prototypes)

    def _find_classwise_prototype(self, matrix: np.ndarray, prototypes: list[int]) -> int:
        mask = np.in1d(range(matrix.shape[0]), prototypes)
        current_partial_distances = np.minimum.reduce(matrix[mask], axis=0) if prototypes else np.inf
        candidate_distances = np.minimum(matrix[~mask], current_partial_distances).sum(axis=1)
        return np.where(~mask)[0][np.argmin(candidate_distances)]

    def _find_single_class_prototypes(self, matrix: np.ndarray) -> list[int]:
        prototypes = []
        for _ in range(self.n_prototypes):
            prototypes.append(
                self._find_classwise_prototype(matrix, prototypes)
            )
        return prototypes

    def select_prototypes(self, x: TDataBatch, y: TTarget | None = None) -> TPrototypes:
        y = y or self.model.get_model_predictions(x)

        classes = self.get_classes(y)
        prototypes = {cls: [] for cls in classes}
        distances = self._create_distance_matrices(x, y, classes)

        for cls in classes:
            class_x = get_x_belonging_to_cls(x, y, cls)
            indices = self._find_single_class_prototypes(distances[cls])
            prototypes[cls] = iloc(class_x, indices)
        return prototypes


class SM_A(IExplainer):
    def __init__(self, model: TModel, n_prototypes: int = 3, *args, **kwargs) -> None:
        super().__init__(model, *args, **kwargs)
        self.n_prototypes = int(n_prototypes)

    def _find_classwise_prototype(self, matrix: np.ndarray, prototypes: list[int]) -> tuple[int, float]:
        if not prototypes:
            idx = np.argmin(matrix.sum(axis=1))
            improvement = matrix.shape[0] - np.sum(matrix[idx])
            return idx, improvement

        mask = np.in1d(range(matrix.shape[0]), prototypes)
        current_partial_distances = matrix[mask].min(axis=0)
        original_distance = current_partial_distances.sum()
        candidates = matrix[~mask]
        candidate_distances = np.minimum(candidates, current_partial_distances).sum(axis=1)
        if not len(candidate_distances):
            return -1, 0.0
        idx = np.argmin(candidate_distances)
        improvement = original_distance - candidate_distances[idx]
        original_idx = np.where(~mask)[0][np.argmin(candidate_distances)]
        return original_idx, improvement

    def _find_prototype(self, distances: dict[int | str, np.ndarray], prototypes: TPrototypes) -> tuple[int | float, int]:
        prototype: tuple[float, int | str, int] = (-np.inf, -1, -1)
        for cls in distances:
            idx, improvement = self._find_classwise_prototype(distances[cls], prototypes[cls])
            if improvement > prototype[0]:
                prototype = (improvement, cls, idx)
        return prototype[1], prototype[2]

    def select_prototypes(self, x: TDataBatch, y: TTarget | None = None) -> TPrototypes:
        y = y or self.model.get_model_predictions(x)

        classes = self.get_classes(y)
        prototypes = {cls: [] for cls in classes}

        distances = self._create_distance_matrices(x, y, classes)
        for _ in range(self.n_prototypes):
            cls, idx = self._find_prototype(distances, prototypes)
            prototypes[cls].append(idx)

        for cls in classes:
            prototypes[cls] = iloc(get_x_belonging_to_cls(x, y, cls), prototypes[cls])

        return prototypes


class SM_WA(SM_A):
    def __init__(self, model: TModel, n_prototypes: int = 3, *args, **kwargs) -> None:
        super().__init__(model=model, n_prototypes=n_prototypes, *args, **kwargs)

    def _find_prototype(self, distances: dict[int | str, np.ndarray], prototypes: dict[int | str, list[int]]
                        ) -> tuple[int | float, int]:
        prototype: tuple[float, int | str, int] = (-np.inf, -1, -1)
        for cls in distances:
            idx, improvement = self._find_classwise_prototype(distances[cls], prototypes[cls])
            improvement = (1 / distances[cls].shape[0]) * improvement
            if improvement > prototype[0]:
                prototype = (improvement, cls, idx)
        return prototype[1], prototype[2]


class SG(SM_A):
    def __init__(self, n_prototypes: int = 3, *args, **kwargs) -> None:
        super().__init__(n_prototypes=n_prototypes, *args, **kwargs)

    def _find_prototype(self, prototypes: TPrototypes, x: TDataBatch, y: TTarget, accuracy: float
                        ) -> tuple[int | str, int, float]:
        prototype: tuple[int | str, int] = (-1, -1)
        for cls in prototypes:
            class_x = x[y.values == cls]
            for idx in range(len(class_x)):
                if idx not in prototypes[cls]:
                    temp_prototypes = prototypes.copy()
                    temp_prototypes[cls].append(idx)
                    for _cls in temp_prototypes:
                        temp_prototypes[_cls] = iloc(get_x_belonging_to_cls(x, y, cls), temp_prototypes[_cls])
                    temp_accuracy = self.score_with_prototypes(x, y, temp_prototypes)
                    if temp_accuracy > accuracy:
                        prototype = (cls, idx)
        return prototype[1], prototype[2], accuracy

    def select_prototypes(self, x: TDataBatch, y: TTarget | None = None) -> TPrototypes:
        y = y or self.model.get_model_predictions(x)

        classes = self.get_classes(y)
        prototypes = {cls: [] for cls in classes}
        accuracy = 0

        for _ in range(self.model.n_prototypes):
            cls, idx, accuracy = self._find_prototype(prototypes, x, y, accuracy)
            prototypes[cls].append(idx)

        for cls in classes:
            prototypes[cls] = iloc(get_x_belonging_to_cls(x, y, cls), prototypes[cls])

        return prototypes


class APete(SM_A):
    def __init__(self, model: TModel, alpha: float = 0.05) -> None:
        super().__init__(model)
        self.alpha = float(alpha)

    def _find_prototype(self, distances: dict[int | str, np.ndarray], prototypes: dict[int | str, list[int]]
                        ) -> tuple[int | float, int, float]:
        delta_prim = -np.inf
        prototype_idx = -1
        prototype_cls = -1
        for cls in distances:
            idx, delta = self._find_classwise_prototype(distances[cls], prototypes[cls])
            if delta > delta_prim:
                delta_prim = delta
                prototype_idx = idx
                prototype_cls = cls
        return prototype_cls, prototype_idx, delta_prim

    def select_prototypes(self, x: TDataBatch, y: TTarget | None = None) -> TPrototypes:
        y = y or self.model.get_model_predictions(x)

        prev_improvement = 0.0
        classes = self.get_classes(y)
        prototypes = {cls: [] for cls in classes}

        distances = self._create_distance_matrices(x, y, classes)
        while True:
            cls, idx, improvement = self._find_prototype(distances, prototypes)
            prototypes[cls].append(idx)

            protos = {}
            for cls in classes:
                protos[cls] = iloc(get_x_belonging_to_cls(x, y, cls), prototypes[cls])

            if improvement == 0 or np.abs(prev_improvement - improvement) / improvement <= self.alpha:
                return protos
            prev_improvement = improvement
